@@ -15,8 +15,8 @@
 #include <math.h>
 #include <ctype.h>
 #include <unistd.h>
-#include <syslog.h>
 #include <fcntl.h>
+#include <stdint.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
@@ -53,27 +53,10 @@ static void explain1(const char *arg)
  */
 #define MAX_DIST	(16*1024)
 
-static const double max_percent_value = 0xffffffff;
-
 /* scaled value used to percent of maximum. */
 static void set_percent(__u32 *percent, double per)
 {
-	*percent = (unsigned int) rint(per * max_percent_value);
-}
-
-
-/* Parse either a fraction '.3' or percent '30%
- * return: 0 = ok, -1 = error, 1 = out of range
- */
-static int parse_percent(double *val, const char *str)
-{
-	char *p;
-
-	*val = strtod(str, &p) / 100.;
-	if (*p && strcmp(p, "%"))
-		return -1;
-
-	return 0;
+	*percent = rint(per * UINT32_MAX);
 }
 
 static int get_percent(__u32 *percent, const char *str)
@@ -89,7 +72,7 @@ static int get_percent(__u32 *percent, const char *str)
 
 static void print_percent(char *buf, int len, __u32 per)
 {
-	snprintf(buf, len, "%g%%", 100. * (double) per / max_percent_value);
+	snprintf(buf, len, "%g%%", (100. * per) / UINT32_MAX);
 }
 
 static char *sprint_percent(__u32 per, char *buf)
@@ -170,7 +153,7 @@ static int get_ticks(__u32 *ticks, const char *str)
 }
 
 static int netem_parse_opt(struct qdisc_util *qu, int argc, char **argv,
-			   struct nlmsghdr *n)
+			   struct nlmsghdr *n, const char *dev)
 {
 	int dist_size = 0;
 	struct rtattr *tail;
@@ -323,7 +306,7 @@ static int netem_parse_opt(struct qdisc_util *qu, int argc, char **argv,
 				/* netem option is "1-h" but kernel
 				 * expects "h".
 				 */
-				gemodel.h = max_percent_value - gemodel.h;
+				gemodel.h = UINT32_MAX - gemodel.h;
 
 				if (!NEXT_IS_NUMBER())
 					continue;
@@ -399,7 +382,12 @@ static int netem_parse_opt(struct qdisc_util *qu, int argc, char **argv,
 		} else if (matches(*argv, "rate") == 0) {
 			++present[TCA_NETEM_RATE];
 			NEXT_ARG();
-			if (get_rate64(&rate64, *argv)) {
+			if (strchr(*argv, '%')) {
+				if (get_percent_rate64(&rate64, *argv, dev)) {
+					explain1("rate");
+					return -1;
+				}
+			} else if (get_rate64(&rate64, *argv)) {
 				explain1("rate");
 				return -1;
 			}
@@ -434,8 +422,6 @@ static int netem_parse_opt(struct qdisc_util *qu, int argc, char **argv,
 		}
 	}
 
-	tail = NLMSG_TAIL(n);
-
 	if (reorder.probability) {
 		if (opt.latency == 0) {
 			fprintf(stderr, "reordering not possible without specifying some delay\n");
@@ -464,8 +450,7 @@ static int netem_parse_opt(struct qdisc_util *qu, int argc, char **argv,
 		return -1;
 	}
 
-	if (addattr_l(n, 1024, TCA_OPTIONS, &opt, sizeof(opt)) < 0)
-		return -1;
+	tail = addattr_nest_compat(n, 1024, TCA_OPTIONS, &opt, sizeof(opt));
 
 	if (present[TCA_NETEM_CORR] &&
 	    addattr_l(n, 1024, TCA_NETEM_CORR, &cor, sizeof(cor)) < 0)
@@ -524,7 +509,7 @@ static int netem_parse_opt(struct qdisc_util *qu, int argc, char **argv,
 			return -1;
 		free(dist_data);
 	}
-	tail->rta_len = (void *) NLMSG_TAIL(n) - (void *) tail;
+	addattr_nest_compat_end(n, tail);
 	return 0;
 }
 
@@ -630,7 +615,7 @@ static int netem_print_opt(struct qdisc_util *qu, FILE *f, struct rtattr *opt)
 		fprintf(f, " loss gemodel p %s",
 			sprint_percent(gemodel->p, b1));
 		fprintf(f, " r %s", sprint_percent(gemodel->r, b1));
-		fprintf(f, " 1-h %s", sprint_percent(max_percent_value -
+		fprintf(f, " 1-h %s", sprint_percent(UINT32_MAX -
 						     gemodel->h, b1));
 		fprintf(f, " 1-k %s", sprint_percent(gemodel->k1, b1));
 	}
